<?php

namespace Drupal\smartlinker_ai\Plugin\AICKEditor;

use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\ai\AiProviderPluginManager;
use Drupal\ai_ckeditor\AiCKEditorPluginBase;
use Drupal\ai_ckeditor\Attribute\AiCKEditor;
use Drupal\ai_ckeditor\Command\AiRequestCommand;
use Drupal\node\NodeInterface;
use Drupal\smartlinker_ai\Service\OpenAiKeyService;
use GuzzleHttp\Exception\RequestException;
use Psr\Http\Client\ClientInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Plugin to Identify internal links and embed them in the selected text.
 */
#[AiCKEditor(
  id: 'smartlinker_ai_identifier',
  label: new TranslatableMarkup('Generate internal links'),
  description: new TranslatableMarkup('Identify internal links and embed them in the selected text'),
)]
final class SmartLinkerAI extends AiCKEditorPluginBase {

  /**
   * The provider plugin manager.
   *
   * @var \Drupal\ai\AiProviderPluginManager
   */
  protected AiProviderPluginManager $aiProviderManager;

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountProxyInterface
   */
  protected $account;

  /**
   * The request stack.
   *
   * @var \Symfony\Component\HttpFoundation\RequestStack
   */
  protected $requestStack;

  /**
   * The logger service.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $loggerFactory;

  /**
   * The HTTP client.
   *
   * @var \Psr\Http\Client\ClientInterface
   */
  protected ClientInterface $httpClient;

  /**
   * The openai key service.
   *
   * @var \Drupal\smartlinker_ai\Service\OpenAiKeyService
   */
  protected OpenAiKeyService $openAiKeyService;

  /**
   * Constructs a InternalLinkIdentifier object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\ai\AiProviderPluginManager $ai_provider_manager
   *   The provider plugin manager.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity storage.
   * @param \Drupal\Core\Session\AccountProxyInterface $account
   *   The instantiated account.
   * @param \Symfony\Component\HttpFoundation\RequestStack $requestStack
   *   A request stack symfony instance.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_factory
   *   The logger factory.
   * @param \Psr\Http\Client\ClientInterface $http_client
   *   The HTTP client.
   * @param \Drupal\smartlinker_ai\Service\OpenAiKeyService $openai_key_service
   *   The openai key service.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, AiProviderPluginManager $ai_provider_manager, EntityTypeManagerInterface $entity_type_manager, AccountProxyInterface $account, RequestStack $requestStack, LoggerChannelFactoryInterface $logger_factory, ClientInterface $http_client, OpenAiKeyService $openai_key_service) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $ai_provider_manager, $entity_type_manager, $account, $requestStack, $logger_factory);
    $this->httpClient = $http_client;
    $this->openAiKeyService = $openai_key_service;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('ai.provider'),
      $container->get('entity_type.manager'),
      $container->get('current_user'),
      $container->get('request_stack'),
      $container->get('logger.factory'),
      $container->get('http_client'),
      $container->get('smartlinker_ai.openai_key_service'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'internal_link_content_type' => [],
      'provider' => NULL,
      'search_index' => NULL,
      'linker_prompt' => '',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $db_options = [];

    /** @var \Drupal\search_api\IndexInterface $index */
    foreach ($this->entityTypeManager->getStorage('search_api_index')->loadMultiple() as $index) {
      $backend = $index->hasValidServer() ? $index->getServerInstance()->getBackendId() : NULL;

      if ($backend == 'search_api_ai_search') {
        $db_options[$index->id()] = $index->label() . ' (' . $index->id() . ')';
      }
    }

    $form['search_index'] = [
      '#type' => 'select',
      '#title' => $this->t('Select the search api index to use'),
      "#empty_option" => $this->t('-- Select --'),
      '#options' => $db_options,
      '#description' => $this->t('Please configure Smartlinker AI server Use vector DB (eg. Milvus, Pinecone) as database and Smartlinker AI Index to select content types to be indexed. . See the <a href=":link">Search Api</a>. Once saved, select the index here.', [':link' => '/admin/config/search/search-api']),
      '#default_value' => $this->configuration['search_index'] ?? '',
    ];

    $options = $this->aiProviderManager->getSimpleProviderModelOptions('chat');
    array_shift($options);
    array_splice($options, 0, 1);
    $form['provider'] = [
      '#type' => 'select',
      "#empty_option" => $this->t('-- Default from AI module (chat) --'),
      '#title' => $this->t('AI provider'),
      '#options' => $options,
      '#default_value' => $this->configuration['provider'] ?? $this->aiProviderManager->getSimpleDefaultProviderOptions('chat'),
      '#description' => $this->t('Select which provider to use for this plugin. See the <a href=":link">Provider overview</a> for details about each provider.', [':link' => '/admin/config/ai/providers']),
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function validateConfigurationForm(array &$form, FormStateInterface $form_state) {

  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $this->configuration['provider'] = $form_state->getValue('provider');
    $this->configuration['internal_link_content_type'] = $form_state->getValue('internal_link_content_type');
    $this->configuration['search_index'] = $form_state->getValue('search_index');
  }

  /**
   * {@inheritdoc}
   */
  public function buildCkEditorModalForm(array $form, FormStateInterface $form_state, array $settings = []) {

    // Attach the module library.
    $form['#attached']['library'][] = 'smartlinker_ai/smartlinker_ai';

    // Get the storage.
    $storage = $form_state->getStorage();

    // Get the editor ID.
    $editor_id = $this->requestStack->getParentRequest()->get('editor_id');

    // Get the search index.
    $search_index = $this->configuration['search_index'];
    if (empty($search_index)) {
      return [
        '#markup' => '<p>' . $this->t('Please set the search index on the respective text editor setting page. <a href=":link" target="_blank">goto setting</a>', [':link' => '/admin/config/content/formats']) . '</p>',
      ];
    }

    // Get the content types.
    $content_options = $this->getSelectedContentTypes($search_index);
    // Check if any content type is selected.
    $field_settings = $this->entityTypeManager->getStorage('search_api_index')->load($search_index)->get('field_settings');
    $type_exists = $this->checkPropertyPath($field_settings);

    $form['content_type'] = [
      '#type' => 'select',
      '#title' => $this->t('Choose the content type from which the links will be attached (optional)'),
      '#options' => $content_options,
      '#description' => $this->t('Choose one if you wish to attach links only to a specific content type. By default, all content types are selected. For example, to attach links only to your case studies, select the "Case Studies" content type.'),
      '#empty_option' => $this->t('All content types'),
      '#attributes' => [
        'class' => ['smartlinker-ai-content-type'],
      ],
      // Hide the field if no content type is selected.
      '#access' => $type_exists,
    ];

    // Number of links to attach.
    $form['links'] = [
      '#type' => 'number',
      '#title' => $this->t('Number of links to attach'),
      '#description' => $this->t('Enter the number of links to attach to the selected text. Please note that if AI does not find the required number of matching internal links, it might either not add the required number of links OR it might modify the results to match this number.'),
      '#step' => 1,
      '#min' => 1,
      '#max' => 30,
    ];

    // Check if the selected text is empty.
    if (empty($storage['selected_text'])) {
      return [
        '#markup' => '<p>' . $this->t('You must select some text before you can attach internal links.') . '</p>',
      ];
    }

    $form = parent::buildCkEditorModalForm($form, $form_state);

    // Keywords to focus on while attaching links.
    $form['keywords'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Keywords'),
      '#description' => $this->t('Add keywords you want AI to focus on, separated by commas (Ex: Keyword1, Keyword2,...)'),
      '#rows' => 1,
    ];

    // Selected text.
    $form['selected_text'] = [
      '#type' => 'textarea',
      '#title' => $this->t('Selected text to attach links'),
      '#disabled' => TRUE,
      '#default_value' => $storage['selected_text'],
    ];

    // Response text.
    $form['response_text'] = [
      '#type' => 'text_format',
      '#title' => $this->t('Suggested text with internal links'),
      '#description' => $this->t('The response from AI will appear in the box above. You can edit and tweak the response before saving it back to the main editor.'),
      '#prefix' => '<div id="ai-ckeditor-response">',
      '#suffix' => '</div>',
      '#default_value' => '',
      '#allowed_formats' => [$editor_id],
      '#format' => $editor_id,
      '#attributes' => ['class' => ['hide-text-format-info']],
    ];

    // Response help text.
    $form['response_text_help'] = [
      '#type' => 'markup',
      '#markup' => '<p>' . $this->t('Please note: Larger text may take up to 20 seconds to process. For best results, we recommend selecting smaller chunks of text.') . '</p>',
    ];

    // Submit button.
    $form['actions']['generate']['#value'] = $this->t('Generate internal links');
    $form['actions']['generate']['#attributes']['class'][] = 'generate-internal-links';

    return $form;
  }

  /**
   * Generate text callback.
   *
   * @param array $form
   *   The form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   *
   * @return mixed
   *   The result of the AJAX operation.
   */
  public function ajaxGenerate(array &$form, FormStateInterface $form_state) {

    // Get the form values.
    $values = $form_state->getValues();

    // Get the selected content types.
    $node_types = $values['plugin_config']['content_type'];

    // Get the keywords.
    $keywords = $values['plugin_config']['keywords'] ?? '';

    $selected_type = '';
    if ($node_types) {
      $content_types = $this->entityTypeManager->getStorage('node_type')->load($node_types);
      $selected_type = $content_types->label();
    }
    $selected_text = $values['plugin_config']['selected_text'];

    $prompt = 'System/Role Instruction:\n';
    $prompt .= 'You are a summarization assistant. I will provide a large body of text. Please create a concise, factual summary that retains the main ideas, domain-specific keywords, and key data points. Do not include subjective commentary or analysis. Limit the summary to around [10 - 20 words per paragraph].\n';
    $prompt .= 'User Prompt: *Summarize the following text for use in a semantic search engine. Make sure to:\n';
    $prompt .= '1. Include important domain-specific terms and key points,\n';
    $prompt .= '2. Avoid extraneous or repetitive details,\n';
    $prompt .= '3. Keep the final summary under [10 - 20 words per paragraph].\n';
    $prompt .= 'Text: ' . $selected_text;

    $content_keywords = $this->getKeywordsFromOpenAi($prompt);

    $links = $this->queryVectorDb($content_keywords, $selected_type, $keywords);
    try {
      // Prompt for the AI CKEditor config.
      $prompt = 'I want you to embed as many relevant links as possible to appropriately matching keywords in the selected content by following the below steps and guidelines\n';
      $prompt .= 'Here is the selected content:\n' . $selected_text;
      $prompt .= '\nThese are the available links and their respective titles that were shared by Milvus, a vector database. The URLs are in json format.:\n' . $links;
      $prompt .= '\n**Step 1: Identify the key titles and URLs you want to link to**\n';
      $prompt .= '**Step 2: Find Strategic Anchor Text**\n';
      $prompt .= '- Once you know what you\'re linking to, find natural places in the content to insert links.\n';
      $prompt .= '- Do not generate any new text or alter the meaning of sentences.\n';
      $prompt .= '- Only hyperlink words or phrases that already exist in the content and are relevant to the linked page\n';
      $prompt .= '- Use each URL only once (no repetition).\n';
      $prompt .= '**Final Output Format:**\n';
      $prompt .= '- Return the updated content with linked keywords.\n';
      $prompt .= '\n **Additional Guidelines:** \n';
      $prompt .= '- Do not repeat keywords or links.\n';
      $prompt .= '- Once a keyword has been linked, do not use another link for it.\n';
      $prompt .= '- Keep original selected content and links unchanged.\n';
      $prompt .= '- Do not add links to headings (h1,h2,h3, etc.)\n';
      $prompt .= '- Do not modify, append, or alter the URLs in any way. Use them exactly as provided.\n';
      $prompt .= '- Ensure that all URLs are valid and not broken (404 errors). If a URL is invalid, do not use it.\n';

      // Check if links are set.
      if (!empty($values['plugin_config']['links'])) {
        $prompt .= '\n Number of Links to embed: Link a minimum of ' . $values['plugin_config']['links'] . ' links to the selected content.\n';
      }

      // Check if keywords are set.
      if (!empty($values['plugin_config']['keywords'])) {
        $prompt .= 'Make sure to focus on these specific keywords: **' . $values['plugin_config']['keywords'] . '**, in addition to the other keywords you have chosen.\n';
      }
      $prompt .= 'Important Notice: If the provided content already contains existing hyperlinks, ensure that these links remain intact and are not removed or replaced when adding new hyperlinks.';

      // Share the prompt with the AI.
      $response = new AjaxResponse();
      $values = $form_state->getValues();
      $response->addCommand(new AiRequestCommand($prompt, $values['editor_id'], $this->pluginDefinition['id'], 'ai-ckeditor-response'));

      // Return the response.
      return $response;
    }
    catch (\Exception $e) {

      // Log the error.
      $this->logger->error('There was an error in the SmartLinker AI Internal Links plugin for CKEditor.');

      // Return the error message.
      return $form['plugin_config']['response_text']['#value'] = 'There was an error in the SmartLinker AI Internal Links plugin for CKEditor.';
    }
  }

  /**
   * Fetches keywords from OpenAI based on the provided prompt.
   *
   * @param string $prompt
   *   The prompt to send to OpenAI.
   *
   * @return string
   *   The response content or an error message.
   */
  public function getKeywordsFromOpenAi($prompt) {
    $url = 'https://api.openai.com/v1/chat/completions';
    $apiKey = $this->openAiKeyService->getOpenApiKey();
    $data = [
      'model' => 'gpt-4',
      'messages' => [['role' => 'user', 'content' => $prompt]],
      'temperature' => 0,
    ];

    try {
      $response = $this->httpClient->post($url, [
        'headers' => [
          'Content-Type' => 'application/json',
          'Authorization' => 'Bearer ' . $apiKey,
        ],
        'json' => $data,
      ]);

      $result = json_decode($response->getBody()->getContents(), TRUE);

      if (!empty($result['choices'][0]['message']['content'])) {
        return trim($result['choices'][0]['message']['content']);
      }
      return "Error: Unexpected API response format";

    }
    catch (RequestException $e) {
      $this->loggerFactory->get('openai')->error('OpenAI API request failed: @message', ['@message' => $e->getMessage()]);
      return "Error: Unable to fetch response from OpenAI";
    }
  }

  /**
   * Queries the vector db to get relevant data on the provided input.
   *
   * @param string $selected_text
   *   The input string to query the vector db.
   * @param string $content_types
   *   The content types to filter the results.
   * @param string $keywords
   *   Keywords to get the content.
   *
   * @return mixed
   *   An json of results from the vector db.
   */
  public function queryVectorDb($selected_text, $content_types = '', $keywords = '') {
    $combined_items = [];
    $search_index = $this->configuration['search_index'];
    if ($search_index) {
      $prompt = 'Find the most relevent contents for these keywords:' . $selected_text;
      $index = $this->entityTypeManager->getStorage('search_api_index')->load($search_index);
      $query = $index->query([
        'limit' => 50,
      ]);
      $query->setOption('search_api_bypass_access', TRUE);
      $query->keys($prompt);
      // Add a condition to filter by content types.
      if (!empty($content_types)) {
        $query->addCondition('type', $content_types, '=');
      }
      $results = $query->execute();
      $combined_items = $results->getResultItems();
      $content = [];

      if ($keywords) {
        $keyword_promt = 'Find the most relevant content for these keywords: ' . $keywords;
        $keyword_query = $index->query([
          'limit' => 20,
        ]);
        $keyword_query->setOption('search_api_bypass_access', TRUE);
        $keyword_query->keys($keyword_promt);
        // Add a condition to filter by content types.
        if (!empty($content_types)) {
          $keyword_query->addCondition('type', $content_types, '=');
        }
        $keyword_results = $keyword_query->execute();
        $combined_items = array_merge($combined_items, $keyword_results->getResultItems());
      }
      $nodes = array_map("unserialize", array_unique(array_map("serialize", $combined_items)));

      $content = [];
      foreach ($nodes as $key => $row) {
        $content[] = $this->getDataFromEntityReference($key);
      }

      return json_encode($content, JSON_UNESCAPED_SLASHES);
    }
    else {
      return [];
    }
  }

  /**
   * Load the canonical URL of a node from its entity reference string.
   *
   * @param string $entity_reference
   *   The entity reference string (e.g., 'entity:node/560').
   *
   * @return array|null
   *   The canonical URL and title of the node, NULL if the node does not exist.
   */
  public function getDataFromEntityReference($entity_reference) {
    // Extract the node ID from the entity reference string.
    if (preg_match('/entity:node\/(\d+)/', $entity_reference, $matches)) {
      $node_id = $matches[1];
      // Load the node entity.
      $node = $this->entityTypeManager->getStorage('node')->load($node_id);
      if ($node instanceof NodeInterface) {
        // Get the canonical URL of the node.
        $link = Url::fromRoute('entity.node.canonical', ['node' => $node_id], ['absolute' => TRUE])->toString();
        $title = $node->label();
        $node_data = [];
        $node_data = [
          'link' => $link,
          'title' => $title,
        ];
        return $node_data;
      }
    }
    return NULL;
  }

  /**
   * Get the list of content types selected in the search index.
   *
   * @param string $index_id
   *   The ID of the search index.
   *
   * @return array
   *   An array of selected content types.
   */
  public function getSelectedContentTypes($index_id) {
    try {
      /** @var \Drupal\search_api\Entity\Index $index */
      $index = $this->entityTypeManager->getStorage('search_api_index')->load($index_id);
      if ($index) {
        $datasource_settings = $index->get('datasource_settings');
        $indexed_content = $datasource_settings['entity:node']['bundles']['selected'];
        $datasource_default = $datasource_settings['entity:node']['bundles']['default'];

        // Load all content types.
        $all_ctypes = $this->entityTypeManager->getStorage('node_type')->loadMultiple();
        $content_type_options = [];

        if ($datasource_default == FALSE && !empty($indexed_content)) {
          // Only selective content types from index.
          foreach ($indexed_content as $content_type) {
            $content_type_entity = $this->entityTypeManager->getStorage('node_type')->load($content_type);
            if ($content_type_entity) {
              $content_type_options[$content_type_entity->id()] = $content_type_entity->label();
            }
          }
        }
        elseif ($datasource_default == TRUE && !empty($indexed_content)) {
          // All content types selected ignoring some.
          $filtered_ctypes = array_filter($all_ctypes, function ($ctype) use ($indexed_content) {
            return !in_array($ctype->id(), $indexed_content);
          });
          foreach ($filtered_ctypes as $content_type) {
            $content_type_options[$content_type->id()] = $content_type->label();
          }
        }
        else {
          // All content types selected ignoring none.
          foreach ($all_ctypes as $content_type) {
            $content_type_options[$content_type->id()] = $content_type->label();
          }
        }

        return $content_type_options;
      }
    }
    catch (\Exception $e) {
      $this->logger->error('There was an error in the SmartLinker AI. Please check the search index configuration.');
    }
    return [];
  }

  /**
   * Check if content type is set on the search index.
   */
  public function checkPropertyPath($array) {
    foreach ($array as $key => $value) {
      if ($key === "property_path" && $value === "type") {
        return TRUE;
      }
      if (is_array($value)) {
        // Recursively search in subarrays.
        if ($this->checkPropertyPath($value)) {
          return TRUE;
        }
      }
    }
    return FALSE;
  }

}
